from typing import List, Optional, Type, Union, TYPE_CHECKING

from json_schema_for_humans import const

if TYPE_CHECKING:
    from json_schema_for_humans.schema.schema_node import SchemaNode


def get_type_name(schema_node: "SchemaNode") -> Optional[str]:
    """Filter. Return the type of a property taking into account the type of items for array and enum"""

    def _python_type_to_json_type(python_type: Type[Union[str, int, float, bool, list, dict]]) -> str:
        return {
            str: const.TYPE_STRING,
            int: const.TYPE_INTEGER,
            float: const.TYPE_NUMBER,
            bool: const.TYPE_BOOLEAN,
            list: const.TYPE_ARRAY,
            dict: const.TYPE_OBJECT,
            type(None): const.TYPE_NULL,
        }.get(python_type, const.TYPE_STRING)

    def _enum_type(enum_values: List["SchemaNode"]) -> str:
        enum_type_names = [
            _python_type_to_json_type(python_type_name)
            for python_type_name in set(type(v.literal) for v in enum_values)
        ]
        if enum_type_names:
            return f"{const.TYPE_ENUM} (of {' or '.join(enum_type_names)})"

        return const.TYPE_ENUM

    def _add_subtype_if_array(type_name: str):
        if type_name == const.TYPE_ARRAY:
            items = schema_node.array_items_def
            if not items:
                return type_name

            subtype = items.keywords.get(const.TYPE)
            if subtype:
                subtype = subtype.literal
            if const.TYPE_ENUM in items.keywords:
                subtype = _enum_type(items.keywords[const.TYPE_ENUM].array_items)

            if not subtype:
                # Too complex to guess items
                return type_name

            type_name = f"{type_name} of {subtype}"

        return type_name

    if const.TYPE_CONST in schema_node.keywords:
        return const.TYPE_CONST
    if const.TYPE_ENUM in schema_node.keywords:
        return _enum_type(schema_node.keywords[const.TYPE_ENUM].array_items)

    type_node = schema_node.keywords.get(const.TYPE)
    if type_node:
        if isinstance(type_node, str):
            type_names = [type_node]
        elif type_node.array_items:
            type_names = [node.literal for node in type_node.array_items]
        else:
            type_names = [type_node.literal]
    else:
        return None

    type_names = [_add_subtype_if_array(type_name) for type_name in type_names]

    return ", ".join(type_names[:-1]) + (" or " if len(type_names) > 1 else "") + type_names[-1]



from enum import Enum
from io import TextIOWrapper, FileIO
from typing import Union, TextIO

TYPE_ARRAY = "array"
TYPE_BOOLEAN = "boolean"
TYPE_CONST = "const"
TYPE_ENUM = "enum"
TYPE_INTEGER = "integer"
TYPE_NULL = "null"
TYPE_NUMBER = "number"
TYPE_OBJECT = "object"
TYPE_STRING = "string"

DESCRIPTION = "description"
DEFAULT = "default"
EXAMPLES = "examples"
ITEMS = "items"
TYPE = "type"
REF = "$ref"
FORMAT = "format"

MULTIPLE_OF = "multipleOf"
MAXIMUM = "maximum"
EXCLUSIVE_MAXIMUM = "exclusiveMaximum"
MINIMUM = "minimum"
EXCLUSIVE_MINIMUM = "exclusiveMinimum"

LINE_WIDTH = 80


FileLikeType = Union[TextIO, TextIOWrapper, FileIO]


class DocumentationTemplate(Enum):
    FLAT = "flat"
    JS = "js"
    MD = "md"
    MD_NESTED = "md_nested"

    @property
    def result_extension(self) -> str:
        if self in [self.FLAT, self.JS]:
            return "html"
        if self in [self.MD, self.MD_NESTED]:
            return "md"


DEFAULT_TEMPLATE_FILE_NAME = "base.html"

DEFAULT_CSS_FILE_NAME = "schema_doc.css"
DEFAULT_JS_FILE_NAME = "schema_doc.min.js"



from typing import Dict, List, Union
from urllib.parse import quote_plus

import jinja2

from json_schema_for_humans import const, jinja_filters
from json_schema_for_humans.schema.schema_node import SchemaNode


def get_numeric_maximum_restriction(schema_node: SchemaNode, default: str = "N/A") -> str:
    """Filter. Get the text to display about maximum restriction on a numeric type(integer or number)"""
    maximum = schema_node.keywords.get(const.MAXIMUM)
    if maximum:
        maximum = maximum.literal
    exclusive_maximum = schema_node.keywords.get(const.EXCLUSIVE_MAXIMUM)
    if exclusive_maximum:
        exclusive_maximum = exclusive_maximum.literal

    # Fix maximum and exclusive_maximum both there
    if maximum is not None and exclusive_maximum is not None:
        if maximum > exclusive_maximum:
            exclusive_maximum = None
        else:
            maximum = None

    maximum_fragment = default
    if maximum is not None:
        maximum_fragment = f"&le; {maximum}"
    if exclusive_maximum is not None:
        maximum_fragment = f"&lt; {exclusive_maximum}"

    return maximum_fragment


def escape_for_table(example_text: str) -> str:
    """Filter. escape characters('|', '`') in string to be inserted into markdown table"""
    return example_text.translate(str.maketrans({"|": "\\|", "`": "\\`"}))


def get_numeric_minimum_restriction(schema_node: SchemaNode, default: str = "N/A") -> str:
    """Filter. Get the text to display about minimum restriction on a numeric type (integer or number)"""
    minimum = schema_node.keywords.get(const.MINIMUM)
    if minimum:
        minimum = minimum.literal
    exclusive_minimum = schema_node.keywords.get(const.EXCLUSIVE_MINIMUM)
    if exclusive_minimum:
        exclusive_minimum = exclusive_minimum.literal

    # Fix minimum and exclusive_minimum both there
    if minimum is not None and exclusive_minimum is not None:
        if minimum <= exclusive_minimum:
            exclusive_minimum = None
        else:
            minimum = None

    minimum_fragment = default
    if minimum is not None:
        minimum_fragment = f"&ge; {minimum}"
    if exclusive_minimum is not None:
        minimum_fragment = f"&gt; {exclusive_minimum}"

    return minimum_fragment


def generate_table(table: List[List[str]]) -> str:
    """
    Pretty print markdown table using list of rows.
    Assuming first row is header line.
    Ending with empty line for rendering bottom border.
    Each column is str padded to max size string in the column.
    """
    if len(table) == 0:
        return ""

    # compute max length of each column
    max_cell_length: Dict = {}
    for idx_row, row in enumerate(table):
        for idx_col, cell in enumerate(row):
            max_cell_length[idx_col] = max(max_cell_length.get(idx_col, 0), len(cell))

    # generate md table
    output = ""
    for idx_row, row in enumerate(table):
        for idx_col, cell in enumerate(row):
            output += "| " + cell.ljust(max_cell_length[idx_col], " ") + " "
        output += "|\n"
        # add header line
        if idx_row == 0:
            for idx_col, cell in enumerate(row):
                output += "| " + "".ljust(max_cell_length[idx_col], "-") + " "
            output += "|\n"

    # add last empty row
    for cell in max_cell_length.values():
        output += "| " + "".ljust(cell, " ") + " "
    output += "|\n"

    return output


def restrictions_table(schema: SchemaNode) -> List[List[str]]:
    """
    String or numeric restrictions tables
    - min/max length
    - regexp pattern + link to regexp101
    - multipleOf
    - minimum/maximum
    ready to be rendered by generate_table filter
    """
    restrictions = []
    if schema.kw_min_length:
        restrictions.append(["**Min length**", str(schema.kw_min_length.literal)])
    if schema.kw_max_length:
        restrictions.append(["**Max length**", str(schema.kw_max_length.literal)])
    if schema.kw_pattern:
        pattern_code = schema.kw_pattern.literal.replace("|", "\\|")
        pattern_url = quote_plus(schema.kw_pattern.literal)
        example_url = ""
        if len(schema.examples) > 0:
            example_url = "&testString=" + quote_plus(schema.examples[0])
        restrictions.append(
            [
                "**Must match regular expression**",
                f"```{pattern_code}``` [Test](https://regex101.com/?regex={pattern_url}{example_url})",
            ]
        )
    if schema.keywords.get("multipleOf"):
        restrictions.append(["**Multiple of**", str(schema.keywords.get("multipleOf").literal)])
    if schema.keywords.get("minimum") or schema.keywords.get("exclusiveMinimum"):
        restrictions.append(["**Minimum**", str(get_numeric_minimum_restriction(schema))])
    if schema.keywords.get("maximum") or schema.keywords.get("exclusiveMaximum"):
        restrictions.append(["**Maximum**", str(get_numeric_maximum_restriction(schema))])

    if len(restrictions) > 0:
        # add header
        restrictions.insert(0, ["Restrictions", " "])

    return restrictions


def array_items(schema: SchemaNode, title: str) -> List[List[str]]:
    """
    List of array items
    ready to be rendered by generate_table filter
    """
    if not schema.array_items:
        return []
    items = [[title]]
    for i, item in enumerate(schema.array_items):
        item_label = item.name_for_breadcrumbs or f"{title} {i}"
        item_html_id = item.html_id
        items.append([f"[{item_label}](#{item_html_id})"])

    return items


def first_line_fixed(example_text: str, max_length: int = 0) -> str:
    """first_line truncated but replace ` with ' to avoid to have only one ` to avoid issues with jekyll"""
    return jinja_filters.first_line(example_text, max_length).translate(str.maketrans({"`": "'"}))


def array_items_restrictions(schema: SchemaNode) -> List[List[str]]:
    """
    Tuple validation restrictions
    ready to be rendered by generate_table filter
    """
    if not schema.array_items_def and not schema.tuple_validation_items:
        return []

    items_restriction_rules = (
        [schema.array_items_def] if schema.array_items_def else []
    ) + schema.tuple_validation_items
    items_restrictions = [["Each item of this array must be", "Description"]]
    for i, item in enumerate(items_restriction_rules):
        item_label = item.name_for_breadcrumbs or f"Array Item {i}"
        item_html_id = item.html_id
        items_restrictions.append(
            [
                f"[{item_label}](#{item_html_id})",
                escape_for_table(first_line_fixed(item.description or "-", const.LINE_WIDTH)),
            ]
        )

    return items_restrictions


class MarkdownTemplate(object):
    def __init__(self, config):
        self.headings = {}
        self.auto_generated_heading = 0
        self.toc = {}
        self.config = config

    def register_jinja(self, env: jinja2.Environment):
        env.filters["md_get_numeric_minimum_restriction"] = get_numeric_minimum_restriction
        env.filters["md_get_numeric_maximum_restriction"] = get_numeric_maximum_restriction
        env.filters["md_escape_for_table"] = escape_for_table
        env.filters["md_heading"] = self.heading
        env.filters["md_properties_table"] = self.properties_table
        env.filters["md_type_info_table"] = self.type_info_table
        env.filters["md_array_restrictions"] = self.array_restrictions
        env.filters["md_array_items_restrictions"] = array_items_restrictions
        env.filters["md_array_items"] = array_items
        env.filters["md_restrictions_table"] = restrictions_table
        env.filters["md_generate_table"] = generate_table
        env.filters["md_first_line"] = first_line_fixed

        env.globals["md_badge"] = self.badge
        env.globals["md_get_toc"] = self.get_toc

    def heading(self, title: str, depth: int, html_id: Union[bool, str] = False, nested: bool = False) -> str:
        """
        Filter. display heading menu, heading number automatically calculated
        from previous heading and depth provided
        """
        if not html_id:
            self.auto_generated_heading = self.auto_generated_heading + 1
            html_id = f"autogenerated_heading_{self.auto_generated_heading}"
        if not (title and title.strip()):
            title = "Auto generated title"
        else:
            title = title.strip()

        # reset heading depth greater than current depth
        for curDepth in range(
            depth + 1, max(int(heading) for heading in self.headings) + 1 if self.headings else depth + 1
        ):
            self.headings.pop(curDepth, None)

        # compute heading, for each depth get last level, and increment level of asked depth
        heading_numbers = ""
        for curDepth in range(0, depth + 1):
            if curDepth in self.headings:
                if curDepth == depth:
                    self.headings[curDepth] = self.headings[curDepth] + 1
            else:
                self.headings[curDepth] = 1
            if curDepth != 0:
                heading_numbers += f"{self.headings[curDepth]}."

        # markdown menu depth
        menu = "#" * min((depth + 1), 5)
        if nested:
            menu = "<strong>"

        # generate markdown title with anchor (except if depth 0)
        if depth == 0:
            menu += f" {title}"
        else:
            if self.config.template_md_options.get("show_heading_numbers"):
                menu += f' <a name="{html_id}"></a>{heading_numbers} {title}'
            else:
                menu += f' <a name="{html_id}"></a>{title}'

        # store current heading in toc
        toc_menu = f"[{title}](#{html_id})"
        if self.config.template_md_options.get("show_heading_numbers"):
            toc_menu = f"[{heading_numbers} {title}](#{html_id})"
        self.toc[heading_numbers] = {"depth": depth, "menu": toc_menu}

        if nested:
            menu += "</strong>"
        return menu

    def get_toc(self) -> str:
        """
        generate Table Of Content from the heading that has been generated
        """
        if not self.config.show_toc:
            return ""

        toc_str = ""

        second_heading_depth = 0
        for i, heading in enumerate(self.toc.values()):
            # Ignore first heading
            if i == 0:
                continue
            if i == 1:
                second_heading_depth = heading["depth"]

            # Ensure we have no space at first level for TOC to be recognized
            indent = "  " * (heading["depth"] - second_heading_depth)
            toc_str += indent + "- " + heading["menu"] + "\n"

        return toc_str

    @staticmethod
    def format_link(title: str, link: str, tooltip: str = "") -> str:
        """Format a Markdown link"""
        return f"[{title}](#{link} {tooltip})"

    def badge(self, name: str, color: str, value: str = "", show_text: bool = False) -> str:
        """
        Badge as markdown image link if badge_as_image option set otherwise Badge as text
        """
        if self.config.template_md_options.get("badge_as_image") and not show_text:
            value_str = ""
            if value and len(value) > 0:
                value_str = "-" + quote_plus(value)
            name = quote_plus(name)
            color = quote_plus(color)
            return f"![badge](https://img.shields.io/badge/{name}{value_str}-{color})"
        else:
            if value and len(value) > 0:
                return f"[{name}: {value}]"
            return f"[{name}]"

    def properties_table(self, schema: SchemaNode) -> List[List]:
        """
        Generate list of properties ready to be rendered by generate_table filter
        """
        properties = []
        for sub_property in schema.iterate_properties:
            line = []
            # property name
            property_name = "+ " if sub_property.is_required_property else "- "
            property_name += self.format_link(escape_for_table(sub_property.property_name), sub_property.html_id)
            line.append(property_name)
            # pattern
            line.append("Yes" if sub_property.is_pattern_property else "No")
            # type
            line.append(
                "Combination" if jinja_filters.is_combining(sub_property) else escape_for_table(sub_property.type_name)
            )
            # Deprecated
            line.append(
                self.badge("Deprecated", "red") if jinja_filters.deprecated(self.config, sub_property) else "No"
            )
            # Link
            if sub_property.should_be_a_link(self.config):
                line.append(
                    "Same as " + self.format_link(sub_property.links_to.link_name, sub_property.links_to.html_id)
                )
            elif sub_property.refers_to:
                line.append("In " + sub_property.ref_path)
            else:
                line.append("-")

            # title or description
            description = sub_property.description or "-"
            if sub_property.title:
                description = sub_property.title

            line.append(escape_for_table(first_line_fixed(description, const.LINE_WIDTH)))

            properties.append(line)

        if properties:
            # add header
            properties.insert(0, ["Property", "Pattern", "Type", "Deprecated", "Definition", "Title/Description"])

        return properties

    def type_info_table(self, schema: SchemaNode) -> List[List]:
        """
        Schema type info table :
        - type,
        - additional properties,
        - default value,
        - definitions links
        ready to be rendered by generate_table filter
        """
        type_info = []

        schema_type = schema.type_name
        default_value = schema.default_value
        type_info.append(["Type", "`combining`" if jinja_filters.is_combining(schema) else f"`{schema_type}`"])
        if jinja_filters.deprecated(self.config, schema):
            type_info.append(["**Deprecated**", self.badge("Deprecated", "red")])

        type_info.append(["**Additional properties**", self.additional_properties(schema)])
        if schema.default_value:
            type_info.append(["**Default**", f"`{default_value}`"])
        if schema.should_be_a_link(self.config):
            schema_link_name = schema.links_to.link_name
            html_id = schema.links_to.html_id
            type_info.append(["**Same definition as**", f"[{ schema_link_name }](#{ html_id })"])
        elif schema.refers_to:
            type_info.append(["**Defined in**", schema.ref_path])

        return type_info

    def additional_properties(self, schema: SchemaNode) -> str:
        """additional properties badge generation"""
        additional_properties = ""
        for sub_property in schema.iterate_properties:
            if sub_property.is_additional_properties:
                if sub_property.is_additional_properties_schema:
                    html_id = sub_property.html_id
                    should_conform_badge = self.badge("Should-conform", "blue")
                    additional_properties = f'[{should_conform_badge}](#{html_id} "Each additional property must conform to the following schema")'
                    break
                else:
                    badge_any_type = self.badge("Any type", "green", "allowed")
                    additional_properties = f'[{badge_any_type}](# "Additional Properties of any type are allowed.")'
                    break

        if len(additional_properties) == 0:
            if schema.explicit_no_additional_properties:
                badge_not_allowed = self.badge("Not allowed", "red")
                additional_properties = f'[{badge_not_allowed}](# "Additional Properties not allowed.")'
            else:
                badge_allowed = self.badge("Any type", "green", "allowed")
                additional_properties = f'[{badge_allowed}](# "Additional Properties of any type are allowed.")'

        return additional_properties

    def array_restrictions(self, schema: SchemaNode) -> List[List[str]]:
        """
        array restrictions: min/max items, items unicity, additional items, Tuple validation
        ready to be rendered by generate_table filter
        """

        if not self.config.template_md_options.get("show_array_restrictions"):
            return []

        return [
            ["", "Array restrictions"],
            ["**Min items**", str(schema.kw_min_items.literal) if schema.kw_min_items else "N/A"],
            ["**Max items**", str(schema.kw_max_items.literal) if schema.kw_max_items else "N/A"],
            [
                "**Items unicity**",
                "True" if schema.kw_unique_items and schema.kw_unique_items.literal is True else "False",
            ],
            [
                "**Additional items**",
                "True" if schema.kw_additional_items and schema.kw_additional_items.literal is True else "False",
            ],
            [
                "**Tuple validation**",
                "See below"
                if schema.array_items_def
                or schema.tuple_validation_items
                or (schema.kw_contains and schema.kw_contains.literal != {})
                else "N/A",
            ],
        ]



import copy
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Union, cast

from json_schema_for_humans import const
from json_schema_for_humans.schema.schema_keyword import SchemaKeyword
from json_schema_for_humans.templating_utils import get_type_name
from json_schema_for_humans.generation_configuration import GenerationConfiguration

circular_references: Dict["SchemaNode", bool] = {}


class SchemaNode:
    """
    Represents a part of a JSON schema with additional metadata to help with documentation
    """

    def __init__(
        self,
        depth: int,
        file: str,
        path_to_element: List[Union[str, int]],
        html_id: str,
        breadcrumb_name: str = "",
        ref_path="",
        parent: "SchemaNode" = None,
        parent_key: str = None,
        literal: Optional[Union[str, int, bool]] = None,
        keywords: Dict[str, Union["SchemaNode", str, List[str]]] = None,
        array_items: List["SchemaNode"] = None,
        array_items_def: Optional["SchemaNode"] = None,
        tuple_validation_items: Optional[List["SchemaNode"]] = None,
        links_to: "SchemaNode" = None,
        refers_to: "SchemaNode" = None,
        is_displayed: bool = True,
    ):
        """

        :param depth: Number of levels from the root of the schema to this node.
        :param file: Real path to the schema file
        :param path_to_element: Path from the root of the schema to the current element
        :param html_id: HTML ID for the current element. Used for anchor links.
        :param parent: The parent node of which the current node is an array item or keyword
        :param parent_key: If the node is under a keyword of the parent node, that keyword
            Example:
            In the following context
            {
                "patternProperties": {
                    ".*": {
                        "type": string
                    }
                }
            }

            For the node
            {
                "type": string
            }
            The parent_key is ".*"

            For the node
            {
                ".*": {
                    "type": string
                }
            }
            The parent key is "patternProperties"

        :param ref_path: Path of a reference to this element, if any (usually "#/definitions/A name")
        :param literal: If the schema is neither a dict nor an array, it will be kept here
                        Useful for things like description, types, const, enum, etc.
        :param keywords: If the schema is a dict, this will be filled. Otherwise, this stays empty
        :param array_items: If the schema is an array, this will be filled. Otherwise, this stays empty
        :param array_items_def: Definition of the array items that this node can contain
        :param tuple_validation_items: Like array_items_def, but defined by position (i.e the element at position i in
                                       a JSON file respecting the schema must respect the schema at position i of this
                                       array)
        :param links_to: If the same node is documented elsewhere, the other SchemaNode that documents it
        :param refers_to: If there is a $ref, this should contain the SchemaNode object for it
        :param is_displayed: Instructs the templates if this part should be fully documented.
                             If false, the description and a link to the referenced element will be generated instead.
                             If false, refers_to needs to be set
        """
        self.depth = depth
        self.file = file
        self.path_to_element = path_to_element
        self.html_id = html_id or "_".join(path_to_element) or "root"
        self.breadcrumb_name = breadcrumb_name
        self.parent = parent
        self.parent_key = parent_key
        self.ref_path = ref_path
        self.literal = literal
        self.keywords = keywords or {}
        self.array_items = array_items or []
        self.links_to = links_to
        self.refers_to = refers_to
        self.is_displayed = is_displayed
        self._refers_to_merged = None
        self.properties: Dict[str, "SchemaNode"] = {}
        self.additional_properties: Optional["SchemaNode"] = None
        # If True, it means additionalProperties is there and false. If False, additionalProperties is either not set
        # or is set but is not false (depends on self.additional_properties)
        self.no_additional_properties: bool = False
        self.pattern_properties: Dict[str, "SchemaNode"] = {}
        self.array_items_def: Optional["SchemaNode"] = array_items_def
        self.tuple_validation_items: List["SchemaNode"] = tuple_validation_items or []

    @property
    def explicit_no_additional_properties(self) -> bool:
        """Return True if additionalProperties is set and false (to differentiate from not set)"""
        return bool(
            (self.properties or self.pattern_properties)
            and self.no_additional_properties
            and not self.additional_properties
        )

    @property
    def definition_name(self) -> str:
        """The text to display when this node is the title of a section or tab"""
        if self.is_property and self.property_name:
            return self.property_name
        if self.title:
            return self.title
        if self.ref_path:
            return self.ref_path.split("/")[-1]
        return ""

    @property
    def link_name(self) -> str:
        """The text to display when linking to this node from somewhere else in the schema"""
        return self.definition_name or self.html_id

    @property
    def name_for_breadcrumbs(self) -> str:
        return self.definition_name or self.breadcrumb_name

    @property
    def is_property(self) -> bool:
        return bool(self.parent and self.property_name in self.parent.properties.keys())

    @property
    def is_pattern_property(self) -> bool:
        return bool(self.parent and self.property_name in self.parent.pattern_properties.keys())

    @property
    def is_additional_properties(self) -> bool:
        return self.parent_key == SchemaKeyword.ADDITIONAL_PROPERTIES.value

    @property
    def is_a_property_node(self) -> bool:
        return self.is_property or self.is_pattern_property or self.is_additional_properties

    @property
    def is_additional_properties_schema(self) -> bool:
        return self.is_additional_properties and self.literal is not True

    @property
    def iterate_properties(self) -> Iterable["SchemaNode"]:
        if self.properties:
            yield from self.properties.values()

        if self.pattern_properties:
            yield from self.pattern_properties.values()

        if self.additional_properties:
            yield self.additional_properties

    @property
    def required_properties(self) -> List[str]:
        """The required properties for this node"""
        required_properties = self.kw_required
        if not required_properties:
            return []

        return [r.literal for r in required_properties.array_items]

    @property
    def is_required_property(self) -> bool:
        """Check if the current node represents a property and that this property is required by its parent"""
        return self.parent and self.property_name in self.parent.required_properties

    @property
    def nodes_from_root(self) -> Iterator["SchemaNode"]:
        """The list of nodes to reach this node"""
        nodes: List["SchemaNode"] = [self]
        current_node = self
        while current_node.parent:
            nodes.append(current_node.parent)
            current_node = current_node.parent

        if len(nodes) == 1:
            # Don't want to display "root" alone at the root
            return []

        return reversed(nodes)

    @property
    def path_to_property(self) -> str:
        """Human-readable representation of the path from the root of the schema to this node"""
        path_without_properties = [
            p
            for p in self.path_to_element
            if p not in [SchemaKeyword.PROPERTIES.value, SchemaKeyword.PATTERN_PROPERTIES.value]
        ]
        return " -> ".join([p if isinstance(p, str) else f"Item {p}" for p in path_without_properties])

    @property
    def flat_path(self) -> str:
        """String representation of the path to this node from the root of the current schema"""
        return "/".join(str(part) for part in self.path_to_element)

    @property
    def default_value(self) -> Optional[Any]:
        def _default_value(node: SchemaNode) -> Optional[Any]:
            default = node.keywords.get(const.DEFAULT)
            if isinstance(default, SchemaNode) and default.is_a_property_node:
                return None
            return default

        seen = set()
        current_node = self
        possible_default = _default_value(current_node)
        while not possible_default and current_node.refers_to:
            if current_node in seen:
                break
            seen.add(current_node)
            current_node = current_node.refers_to
            possible_default = _default_value(current_node)

        return possible_default

    @property
    def format(self) -> Optional[str]:
        format_val = self.keywords.get(const.FORMAT)
        if format_val:
            return format_val.literal
        return None

    @property
    def description(self) -> str:
        description = ""
        description_node = self.keywords.get(const.DESCRIPTION)
        if description_node:
            description = description_node.literal

        seen = set()
        current_node = self
        while not description and current_node.refers_to:
            if current_node in seen:
                break
            seen.add(current_node)
            referenced_schema = current_node.refers_to
            referenced_description_node = referenced_schema.keywords.get(const.DESCRIPTION)
            if referenced_description_node:
                description = referenced_description_node.literal
            current_node = referenced_schema

        return description

    @property
    def examples(self) -> List[str]:
        possible_examples = self.keywords.get(const.EXAMPLES)
        if not possible_examples:
            return []

        if isinstance(possible_examples, SchemaNode) and possible_examples.is_a_property_node:
            return []

        return possible_examples

    @property
    def refers_to_merged(self) -> Optional["SchemaNode"]:
        """The referenced node, with values from the current node merged in"""
        if self._refers_to_merged:
            return self._refers_to_merged

        if not self.refers_to:
            return None

        merged_node = copy.copy(self.refers_to)
        merged_node.keywords = {k: copy.copy(v) for k, v in self.refers_to.keywords.items()}
        merged_node.array_items = [copy.copy(i) for i in self.refers_to.array_items]

        merged_node.keywords.update({k: copy.copy(v) for k, v in self.keywords.items()})
        merged_node.array_items += [copy.copy(i) for i in self.array_items]

        return merged_node

        # self._refers_to_merged = merged_node

        # return self._refers_to_merged

    def get_keyword(self, keyword: SchemaKeyword) -> Optional["SchemaNode"]:
        """Get the value of a keyword if present and it is not a property (to avoid conflicts with properties being
        named like a keyword, e.g. a property named "if")
        """
        possible_keyword = self.keywords.get(keyword.value)
        if possible_keyword and isinstance(possible_keyword, SchemaNode) and not possible_keyword.is_property:
            return possible_keyword

        return None

    @property
    def kw_all_of(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ALL_OF)

    @property
    def kw_any_of(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ANY_OF)

    @property
    def kw_one_of(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ONE_OF)

    @property
    def kw_not(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.NOT)

    @property
    def has_conditional(self) -> bool:
        return self.kw_if is not None and (self.kw_then is not None or self.kw_else is not None)

    @property
    def kw_if(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.IF)

    @property
    def kw_then(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.THEN)

    @property
    def kw_else(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ELSE)

    @property
    def kw_enum(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ENUM)

    @property
    def kw_const(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.CONST)

    @property
    def kw_pattern(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.PATTERN)

    @property
    def kw_properties(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.PROPERTIES)

    @property
    def kw_pattern_properties(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.PATTERN_PROPERTIES)

    @property
    def kw_additional_properties(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ADDITIONAL_PROPERTIES)

    @property
    def kw_min_length(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.MIN_LENGTH)

    @property
    def kw_max_length(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.MAX_LENGTH)

    @property
    def kw_items(self) -> Optional[List["SchemaNode"]]:
        """items can be either an object either a list of object"""
        items = self.get_keyword(SchemaKeyword.ITEMS)
        if not items:
            return None

        if items.array_items:
            return items.array_items

        return [items]

    @property
    def kw_min_items(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.MIN_ITEMS)

    @property
    def kw_max_items(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.MAX_ITEMS)

    @property
    def kw_unique_items(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.UNIQUE_ITEMS)

    @property
    def kw_additional_items(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.ADDITIONAL_ITEMS)

    @property
    def kw_contains(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.CONTAINS)

    @property
    def kw_required(self) -> Optional["SchemaNode"]:
        return self.get_keyword(SchemaKeyword.REQUIRED)

    @property
    def title(self) -> Optional[str]:
        title_kw = self.get_keyword(SchemaKeyword.TITLE)
        if not title_kw:
            return None
        title = title_kw.literal
        title = cast(str, title)
        return title

    @property
    def property_name(self) -> Optional[str]:
        return self.parent_key

    @property
    def property_display_name(self) -> Optional[str]:
        """The name to display in documentation for this property.

        This is simply the property name unless it is under "patternProperties" and it has a title,
        in which case it is that title
        """
        if self.is_pattern_property:
            return self.title or self.parent_key
        if self.is_additional_properties:
            return "Additional Properties"
        return self.parent_key

    @property
    def type_name(self) -> str:
        name = get_type_name(self)

        if name:
            return name

        seen = set()
        current_node = self
        while not name and current_node.refers_to:
            if current_node in seen:
                break
            seen.add(current_node)
            referenced_schema = current_node.refers_to
            name = get_type_name(referenced_schema)
            current_node = referenced_schema

        return name or const.TYPE_OBJECT

    @property
    def raw(self) -> Optional[Union[int, bool, str, List, Dict]]:
        """Get the value of the node as it would exist in the original schema.
        Useful for const where we don't care for the structure and just want to display the value
        """
        if self.literal is not None:
            return self.literal
        if self.array_items:
            return [node.raw for node in self.array_items]
        if self.keywords:
            return {k: v.raw for k, v in self.keywords.items()}
        return None

    def should_be_a_link(self, config: GenerationConfiguration) -> bool:
        """Check if this node should be displayed as a link to another section of the schema in the context of
        the provided configuration.
        """
        if not self.links_to or self.is_displayed:
            return False

        if config.link_to_reused_ref:
            return True

        return self.has_circular_reference(config)

    def node_is_parent(self, node_to_check: "SchemaNode") -> bool:
        """Check if the provided node is a parent of the current node"""
        if self.file != node_to_check.file:
            return False

        for i, path_part in enumerate(node_to_check.path_to_element):
            if len(self.path_to_element) <= i:
                return False
            if self.path_to_element[i] != path_part:
                return False
        return True

    def has_circular_reference(self, config: GenerationConfiguration) -> bool:
        """Check if the current schema is a reference to another section that references the current schema.

        The check is recursive up to config.recursive_detection_depth levels, meaning that if the node refers to another
        node that refers to another node that refers to a parent of itself, this will still return True if, and only if,
        it takes less than config.recursive_detection_depth steps to get to the parent.
        """
        if self in circular_references:
            return circular_references[self]

        if not self.links_to:
            circular_references[self] = False
            return False

        iteration_count = 0
        to_check = {self.links_to}
        while to_check and iteration_count < config.recursive_detection_depth:
            for node_to_check in to_check:
                # If the node reached via reference, keywords, or array items is the node itself, we have a circular
                # reference.
                # We also check if the path is for a parent to save on cycles
                if node_to_check == self or self.node_is_parent(node_to_check):
                    circular_references[self] = True
                    return True

            new_to_check: Set[SchemaNode] = set()
            for node_to_check in to_check:
                if node_to_check.links_to:
                    new_to_check.add(node_to_check.links_to)
                new_to_check = new_to_check.union(
                    set(n for n in node_to_check.keywords.values() if isinstance(n, SchemaNode))
                )
                new_to_check = new_to_check.union(set(node_to_check.array_items))
            to_check = new_to_check
            iteration_count += 1

        circular_references[self] = False
        return False

    def __eq__(self, other: object) -> bool:
        """For two schema nodes to be considered equals they must represent the same element in the same file"""
        if other is None:
            return False

        if not isinstance(other, SchemaNode):
            return NotImplemented

        return self.file == other.file and self.path_to_element == other.path_to_element

    def __hash__(self) -> int:
        return hash(self.file + self.flat_path)

    def __str__(self) -> str:
        return self.flat_path



import json
import logging
import os
from dataclasses import dataclass
from json import JSONDecodeError
from pathlib import Path
from typing import Any, Union, Dict, List, Optional

import yaml
from dataclasses_json import dataclass_json

from json_schema_for_humans.const import (
    DocumentationTemplate,
    FileLikeType,
    DEFAULT_CSS_FILE_NAME,
    DEFAULT_JS_FILE_NAME,
)


@dataclass_json
@dataclass
class GenerationConfiguration:
    """Configuration for generating documentation for a schema"""

    minify: bool = True
    description_is_markdown: bool = True
    deprecated_from_description: bool = False
    show_breadcrumbs: bool = True
    collapse_long_descriptions: bool = True
    default_from_description: bool = False
    expand_buttons: bool = False
    copy_css: bool = True
    copy_js: bool = True
    link_to_reused_ref: bool = True
    recursive_detection_depth: int = 25
    templates_directory: Union[str, Path] = Path(__file__).parent / "templates"
    template_name: str = "js"
    custom_template_path: Optional[str] = None
    show_toc: bool = True
    examples_as_yaml: bool = False
    # markdown2 extra parameters can be added here: https://github.com/trentm/python-markdown2/wiki/Extras
    markdown_options: Optional[Dict[str, Any]] = None
    template_md_options: Optional[Dict[str, Any]] = None
    with_footer: bool = True
    footer_show_time: bool = True

    def __post_init__(self) -> None:
        default_markdown_options = {
            "break-on-newline": True,
            "fenced-code-blocks": {"cssclass": "highlight jumbotron"},
            "tables": None,
        }
        default_markdown_options.update(self.markdown_options or {})
        self.markdown_options = default_markdown_options

        default_template_md_options = {
            "badge_as_image": False,
            "show_heading_numbers": True,
            "show_array_restrictions": True,
        }
        default_template_md_options.update(self.template_md_options or {})
        self.template_md_options = default_template_md_options

    @property
    def templates_directory_path(self) -> Path:
        return Path(self.templates_directory)

    @property
    def files_to_copy(self) -> List[str]:
        files_to_copy: List[str] = []
        if self.copy_js:
            files_to_copy.append(DEFAULT_JS_FILE_NAME)
        if self.copy_css:
            files_to_copy.append(DEFAULT_CSS_FILE_NAME)
        return files_to_copy

    @property
    def template_is_markdown(self) -> bool:
        return self.result_extension == "md"

    @property
    def template_is_html(self) -> bool:
        return self.result_extension == "html"

    @property
    def result_extension(self) -> str:
        """File extension for the resulting documentation"""
        if self.custom_template_path:
            return Path(self.custom_template_path).suffix[1:]

        if self.template_name:
            return DocumentationTemplate(self.template_name).result_extension

        raise ValueError("Trying to get extension for configuration with no template")

    @property
    def template_path(self) -> Optional[Path]:
        if self.custom_template_path:
            return Path(self.custom_template_path)

        return (
            Path(__file__).parent
            / "templates"
            / self.template_name
            / f"base.{DocumentationTemplate(self.template_name).result_extension}"
        )


CONFIG_DEPRECATION_MESSAGE = (
    "JSON Schema for humans: Please supply a GenerationConfiguration object instead of individual options"
)


def get_final_config(
    minify: bool,
    deprecated_from_description: bool,
    default_from_description: bool,
    expand_buttons: bool,
    link_to_reused_ref: bool,
    copy_css: bool = False,
    copy_js: bool = False,
    config: Optional[Union[str, Path, FileLikeType, Dict[str, Any], GenerationConfiguration]] = None,
    config_parameters: List[str] = None,
) -> GenerationConfiguration:
    if config:
        final_config = _load_config(config)
    else:
        final_config = GenerationConfiguration(
            minify=minify,
            deprecated_from_description=deprecated_from_description,
            default_from_description=default_from_description,
            expand_buttons=expand_buttons,
            link_to_reused_ref=link_to_reused_ref,
            copy_css=copy_css,
            copy_js=copy_js,
        )
        if (
            not minify
            or deprecated_from_description
            or default_from_description
            or expand_buttons
            or not link_to_reused_ref
        ):
            logging.info(CONFIG_DEPRECATION_MESSAGE)

    if config_parameters:
        final_config = _apply_config_cli_parameters(final_config, config_parameters)

    return final_config


def _load_config(
    config_parameter: Union[str, Path, FileLikeType, Dict[str, Any], GenerationConfiguration]
) -> GenerationConfiguration:
    """Load the configuration from either the path (as str or Path) to a config file, the open config file object,
    The loaded config as a dict or the GenerateConfiguration object directly.
    """
    if isinstance(config_parameter, GenerationConfiguration):
        return config_parameter

    if isinstance(config_parameter, dict):
        config_dict = config_parameter
    elif isinstance(config_parameter, (str, Path)):
        if isinstance(config_parameter, str):
            real_path = os.path.realpath(config_parameter)
        else:
            real_path = str(config_parameter.resolve())
        with open(os.path.realpath(real_path), encoding="utf-8") as config_fp:
            config_dict = yaml.safe_load(config_fp.read())
    else:
        config_dict = yaml.safe_load(config_parameter.read())

    return GenerationConfiguration.from_dict(config_dict)  # type: ignore


def _apply_config_cli_parameters(
    current_configuration: GenerationConfiguration, config_cli_parameters: List[str]
) -> GenerationConfiguration:
    current_configuration_as_dict = current_configuration.to_dict()  # type: ignore
    parameter_value: Union[str, bool, int]
    for parameter in config_cli_parameters:
        if "=" in parameter:
            parameter_name, parameter_value = parameter.split("=")
            try:
                parameter_value = json.loads(parameter_value)
            except JSONDecodeError:
                pass
        else:
            parameter_name = parameter
            if parameter_name.startswith("no_") or parameter_name.startswith("no-"):
                parameter_value = False
                parameter_name = parameter_name[3:]  # Strip the `no_`/`no-`
            else:
                parameter_value = True
        current_configuration_as_dict[parameter_name] = parameter_value

    return GenerationConfiguration.from_dict(current_configuration_as_dict)  # type: ignore
